package ru.dbotthepony.mc.otm.capability.matter

import net.minecraft.world.item.Item
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import java.util.*
import java.util.function.Predicate
import java.util.stream.Collectors
import java.util.stream.Stream

interface IMatterHandler {
	val storedMatter: ImpreciseFraction
	val maxStoredMatter: ImpreciseFraction

	fun receiveMatterOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction
	fun receiveMatterInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction
	fun extractMatterOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction
	fun extractMatterInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction

	val direction: MatterDirection
	val missingMatter: ImpreciseFraction
		get() = maxStoredMatter.minus(storedMatter).moreThanZero()

	val allowsExtract: Boolean
		get() = direction != MatterDirection.RECEIVE

	val allowsReceive: Boolean
		get() = direction != MatterDirection.EXTRACT
}

enum class MatterDirection { RECEIVE, EXTRACT, BIDIRECTIONAL }

interface IReplicationTaskProvider {
	/**
	 * It must return new stream each time
	 */
	val replicationTasks: Stream<out IReplicationTask<*>>

	/**
	 * It must return new stream each time
	 */
	val allReplicationTasks: Stream<out IReplicationTask<*>>

	/**
	 * Allocates (marks as work-in-progress) a task
	 * by incrementing it's [IReplicationTask.inProgress] by 1
	 * and shrinking [IReplicationTask.required] by 1
	 *
	 * If [IReplicationTask.required] == 0, it should not be returned by this method
	 * @param simulate whenever to change internal state
	 * @return MatterTaskAllocation(task, pattern) that should be performed, or null if no work is available
	 */
	fun allocateTask(simulate: Boolean): ReplicationTaskAllocation?

	/**
	 * Notify about task completion. If this provider indeed contain this task, it should
	 * shrink in_progress by 1
	 * If in_progress == 0 and required == 0, it should discard the task
	 * @param taskId task being completed. this method should ignore tasks that are not owned by it.
	 * @return whenever task indeed belong to this provider and internal state was updated
	 */
	fun notifyTaskCompletion(taskId: UUID): Boolean

	/**
	 * @param id uuid of task
	 * @return MatterTask that this capability holds with this id, or null
	 */
	fun getTask(id: UUID): IReplicationTask<*>? {
		return allReplicationTasks.filter { it.id == id }.findAny().orElse(null)
	}

	/**
	 * Destroys all tasks this capability contains
	 */
	fun dropAllTasks()
}

interface IPatternStorage {
	/**
	 * It must return new stream each time
	 */
	val storedPatterns: Stream<out IPatternState>

	fun findPatterns(item: Item): Collection<IPatternState> {
		return findPatterns { item == it.item }
	}

	fun findPatterns(predicate: Predicate<IPatternState>): Collection<IPatternState> {
		return storedPatterns.filter(predicate).collect(Collectors.toList())
	}

	fun findPattern(item: Item): IPatternState? {
		return storedPatterns.filter { it.item == item }.findAny().orElse(null)
	}

	fun getPattern(id: UUID?): IPatternState? {
		return storedPatterns.filter { it.id == id }.findAny().orElse(null)
	}

	fun hasPattern(item: Item): Boolean {
		return storedPatterns.filter { it.item == item }.findAny().isPresent
	}

	fun hasPattern(state: IPatternState): Boolean {
		return hasPattern(state.id)
	}

	fun hasPattern(id: UUID?): Boolean {
		return storedPatterns.filter { it.id == id }.findAny().isPresent
	}

	val capacity: Int
	val stored: Int

	/**
	 * Prefer way to call it is to first call only_update = true
	 * and if it fail, try only_update = false
	 *
	 * @param pattern pattern to be inserted or update value from
	 * @param onlyUpdate do not insert new pattern if this pattern's UUID is not matched
	 * @param simulate whenever to affect state
	 * @return record of status of the operation (at status() FAIL, UPDATED, INSERTED) as well as new_state and old_state
	 */
	fun insertPattern(pattern: IPatternState, onlyUpdate: Boolean, simulate: Boolean): PatternInsertStatus

	fun insertPattern(pattern: IPatternState, simulate: Boolean): PatternInsertStatus {
		return insertPattern(pattern, false, simulate)
	}

	fun updatePattern(pattern: IPatternState, simulate: Boolean): PatternInsertStatus {
		return insertPattern(pattern, true, simulate)
	}
}
